import asyncio
import logging

from .base import DEFAULT_RECEIVABLE_EVENTS, Vkontakte


class VkontakteLongpoll(Vkontakte):
    def __init__(self, *args, longpoll_settings=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._longpoll_settings = longpoll_settings or {}
        self._longpoll_data = {}
        self._longpoll_url_template = "{}?act=a_check&key={}&wait=25&ts={}"

    async def _update_longpoll_data(self):
        self._longpoll_data = await self._direct_request(
            "groups.getLongPollServer", {"group_id": self.group["id"]}
        )

    async def on_start(self, app):
        await super().on_start(app)

        await self._direct_request(
            "groups.setLongPollSettings",
            {
                "group_id": self.group["id"],
                "api_version": self.api_version,
                "enabled": 1,
                **DEFAULT_RECEIVABLE_EVENTS,
                **self._longpoll_settings,
            },
        )

        await self._update_longpoll_data()

    async def _fetch_updates(self):
        try:
            response = await self.client.post(
                self._longpoll_url_template.format(
                    self._longpoll_data["server"],
                    self._longpoll_data["key"],
                    self._longpoll_data["ts"],
                )
            )
            return response.json()
        except asyncio.CancelledError:
            raise
        except Exception:
            logging.exception("Exceptions while gettings updates (Vkontakte)")
            await asyncio.sleep(1)
            return None

    async def acquire_updates(self, queue: asyncio.Queue):
        while True:
            response = await self._fetch_updates()

            if not response:
                continue

            if "ts" in response:
                self._longpoll_data["ts"] = response["ts"]

            if "failed" in response:
                if response["failed"] in (2, 3):
                    await self._update_longpoll_data()
                continue

            for update in response["updates"]:
                if update["type"] == "group_change_settings":
                    await self._update_group_data()

                await queue.put((self._make_update(update), self))



import asyncio
import logging

from aiohttp import web

from ...helpers import get_random_string
from .base import DEFAULT_RECEIVABLE_EVENTS, Vkontakte


class VkontakteCallback(Vkontakte):
    def __init__(
        self,
        *args,
        address=None,
        path="/",
        host="0.0.0.0",
        port=10888,
        updates_queue_maxsize=0,
        callback_settings=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._address = address
        self._path = path
        self._host = host
        self._port = port
        self._updates_queue_maxsize = updates_queue_maxsize
        self._callback_settings = callback_settings or {}

        self._app_runner = None
        self._updates_queue: asyncio.Queue

    async def _handle_request(self, request):
        data = await request.json()

        if data["type"] == "group_change_settings":
            await self._update_group_data()

        elif data["type"] == "confirmation":
            response = await self._direct_request(
                "groups.getCallbackConfirmationCode",
                {"group_id": data["group_id"]},
            )

            return web.Response(body=f"{response['code']}")

        await self._updates_queue.put(self._make_update(data))

        return web.Response(body="ok")

    def _make_server_app(self):
        app = web.Application()
        app.add_routes([web.post(self._path, self._handle_request)])
        return app

    async def _start_server(self):
        app = self._make_server_app()

        self._app_runner = web.AppRunner(app)

        await self._app_runner.setup()

        site = web.TCPSite(self._app_runner, self._host, self._port)

        await site.start()

    async def _stop_server(self):
        if self._app_runner:
            await self._app_runner.cleanup()

    async def on_start(self, app):
        await super().on_start(app)

        # Prepare queue and start server
        self._updates_queue = asyncio.Queue(self._updates_queue_maxsize)

        await self._start_server()

        # Notify user if we won't setup group
        if not self._address:
            logging.warning(
                "No address provided for VkontakteCallback! You "
                "will have to setup your group's settings manually."
            )
            return

        # Delete other callback servers poiting to our address
        servers = await self._direct_request(
            "groups.getCallbackServers",
            {"group_id": self.group["id"]},
        )

        for server in servers["items"]:
            if server["url"] == self._address:
                await self._direct_request(
                    "groups.deleteCallbackServer",
                    {
                        "group_id": self.group["id"],
                        "server_id": server["id"],
                    },
                )

        # Add a new callback server
        response = await self._direct_request(
            "groups.addCallbackServer",
            {
                "group_id": self.group["id"],
                "url": self._address,
                "title": get_random_string(),
                "secret_key": get_random_string(24),
            },
        )

        server_id = response["server_id"]

        # Updated callback server's settings
        await self._direct_request(
            "groups.setCallbackSettings",
            {
                "group_id": self.group["id"],
                "api_version": self.api_version,
                "server_id": server_id,
                **DEFAULT_RECEIVABLE_EVENTS,
                **self._callback_settings,
            },
        )

    async def on_shutdown(self, app):
        await super().on_shutdown(app)
        await self._stop_server()

    async def acquire_updates(self, queue):
        while True:
            await queue.put((await self._updates_queue.get(), self))



import importlib
import importlib.util
import logging
import os
import pkgutil
import sys

from .plugin import Plugin


def _validate_plugins(plugins, package):
    for plugin in plugins:
        if not isinstance(plugin, Plugin):
            raise ValueError(
                f'Package "{package}" exported {plugin}, which is not an instance of Plugin'
            )

    return plugins


def _extract_plugins(package):
    if hasattr(package, "plugin"):
        logging.info('Loaded 1 plugin from "%s" package', package.__name__)
        return _validate_plugins([package.plugin], package.__name__)

    if hasattr(package, "plugins"):
        logging.info(
            'Loaded %d plugin(s) from "%s" package',
            len(package.plugins),
            package.__name__,
        )
        return _validate_plugins(package.plugins, package.__name__)

    logging.debug('No plugins found in "%s" package', package.__name__)
    return []


def _load_package(path):
    # Acquire finder for provided folder (if provided path is file)
    # or folder's parent (if provided path is folder)
    if os.path.isdir(path):
        parent_path = os.path.join(path, "..")
    else:
        parent_path = os.path.dirname(path)

    finder = pkgutil.get_importer(parent_path)
    if finder is None:
        raise ValueError(f'Failed to get importer from path "{parent_path}"')

    # Extract module name
    _, package_name = os.path.split(path)
    package_name = "".join(package_name.rsplit(".py", 1))

    # Import module using extracted name
    spec = finder.find_spec(package_name)
    if spec is None or spec.loader is None:
        raise ValueError(
            f'Failed to get spec for "{package_name}" (or spec has no loader)'
        )

    module = importlib.util.module_from_spec(spec)
    sys.modules[package_name] = module
    spec.loader.exec_module(module)

    # Return imported module
    return module


def load_plugins_from_module(module):
    return _extract_plugins(module)


def load_plugins_from_path(path):
    """
    Return list of plugins loaded from specified path. Modules are
    loaded recursively, packages are loaded non-recursively. Packages
    are detected using "__init__.py" file. Path itself can be package.
    Only plugins in module-level variables "plugin" or "plugins" are
    loaded.

    :param path: path to target folder
    :returns: list of loaded plugins
    """

    entries = os.listdir(path)

    if "__init__.py" in entries:
        return _extract_plugins(_load_package(path))

    plugins = []

    for entry in entries:
        entry_path = os.path.join(path, entry)

        if entry.startswith("__"):
            continue

        if os.path.isdir(entry_path):
            plugins.extend(load_plugins_from_path(entry_path))
        elif entry_path.endswith(".py"):
            plugins.extend(_extract_plugins(_load_package(entry_path)))

    return plugins



import asyncio

from ..backend import Backend
from ..kutana import Kutana
from ..plugin import Plugin
from ..update import Message, RecipientKind


class Debug(Backend):
    def __init__(self, updates, identity=None, **kwargs):
        super().__init__(**kwargs)

        self.updates = [self._make_update(*args) for args in updates]
        self._update_processed = 0

        self.messages = []
        self.requests = []

        self._identity = identity

    def get_identity(self):
        return self._identity

    @classmethod
    async def handle_updates(cls, plugins, updates, identity=None):
        # create backend
        backend = cls(updates, identity=identity)

        # create plugin that will stop applition when needed
        stopper_plugin = Plugin("_stopper")

        def _update_processed():
            backend._update_processed += 1
            if backend._update_processed >= len(backend.updates):
                future.cancel()

        @stopper_plugin.on_completion()
        async def _(context):
            _update_processed()

        @stopper_plugin.on_exception()
        async def _(context, exception):
            _update_processed()

        # create application, fill it with plugins and backends
        app = Kutana()

        app.add_backend(backend)

        app.add_plugin(stopper_plugin)

        for plugin in plugins:
            app.add_plugin(plugin)

        # run application (save to variable for stopper)
        future = asyncio.ensure_future(app._run())

        try:
            await future
        except asyncio.CancelledError:
            pass
        finally:
            await app._shutdown_wrapper()

        # return used application and backend
        return app, backend

    def _make_update(self, text, sender_id, recipient_id, attachments, raw=None):
        return Message(
            sender_id=sender_id,
            recipient_id=recipient_id,
            recipient_kind=RecipientKind.PRIVATE_CHAT,
            text=text,
            attachments=attachments,
            date=0,
            raw=raw,
        )

    async def acquire_updates(self, queue):
        while self.updates:
            await queue.put((self.updates.pop(0), self))

    async def send_message(self, target_id, message, attachments, **kwargs):
        self.messages.append((target_id, str(message), attachments, kwargs))

    async def request(self, method, kwargs):
        self.requests.append((method, kwargs))



class Symbol:
    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        return f"<Symbol({self.name})>"


PROCESSED = Symbol("PROCESSED")
SKIPPED = Symbol("SKIPPED")



import random
from threading import Lock
from typing import Optional

from ..storage import Document, OptimisticLockException, Storage


class MemoryStorage(Storage):
    """
    Naive implementation of the storage that uses in-memory dict for
    storing data.

    WARNING:
        When keys count is equal or greater than 'keys_limit', 30% of the
        stored keys will be cleared.
    """

    def __init__(self, keys_limit=1_000_000):
        self._keys_limit = keys_limit
        self._storage = {}
        self._lock = Lock()

    def _get_document(self, data, key) -> Optional[Document]:
        if data:
            return Document(data, _storage=self, _storage_key=key)
        return data

    async def put(self, key, data):
        with self._lock:
            # handle keys limit overflow
            if len(self._storage) >= self._keys_limit:
                for key in random.sample(
                    list(self._storage.keys()), k=int(self._keys_limit * 0.3)
                ):
                    self._storage.pop(key)

            # use optimistic locking
            if not data.get("_version") and key in self._storage:
                raise OptimisticLockException("Data for this key already exists")

            if data.get("_version") and key not in self._storage:
                raise OptimisticLockException("Data for this key was deleted")

            if data.get("_version") != self._storage.get(key, {}).get("_version"):
                raise OptimisticLockException(
                    "Data version differ from the expected value"
                )

            # process data and save it
            document = {
                **{k: v for k, v in data.items() if v is not None},
                "_version": (data.get("_version") or 0) + 1,
            }

            self._storage[key] = document

            return self._get_document(document, key)

    async def get(self, key):
        with self._lock:
            return self._get_document(self._storage.get(key, None), key)

    async def delete(self, key):
        with self._lock:
            self._storage.pop(key, None)



import os.path
import random
import string

import httpx


def create_httpx_async_client():
    return httpx.AsyncClient(timeout=55)


def second_arg(_, value):
    return value


def uniq_by(arr, key=None):
    if key:
        return list({key(element): element for element in arr}.values())
    else:
        return list(set(arr))


def pick(obj, keys):
    new_obj = {}
    for key in keys:
        if key in obj:
            new_obj[key] = obj[key]
    return new_obj


def pick_by(obj, predicate=None):
    predicate = predicate or second_arg
    new_obj = {}
    for key, value in obj.items():
        if predicate(key, value):
            new_obj[key] = value
    return new_obj


def chunks(iterable, length=4096):
    for i in range(0, len(iterable), length):
        yield iterable[i : i + length]


def get_random_string(length=14):
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def make_hash(obj, hash_func=hash):
    """
    Return hash of provided object. This method supports hash of:
    - set, tuple, list;
    - dict;
    - other valuse already supported by "hash_func".
    """
    if isinstance(obj, (set, tuple, list)):
        return tuple(map(make_hash, obj))
    elif isinstance(obj, dict):
        return make_hash(tuple(obj.items()))
    else:
        return hash_func(obj)


def get_path(root, path):
    """
    Shortcut for ``os.path.join(os.path.dirname(root), path)``.

    :param root: root path
    :param path: path to file or folder
    :returns: path to file or folder relative to root
    """

    return os.path.join(os.path.dirname(root), path)



import re
from typing import Dict, List, Optional

from .context import Context
from .handler import PROCESSED, SKIPPED
from .update import Message


class Router:
    def __init__(self, priority: Optional[int] = 0):
        self.priority = priority

    def add_handler(self, handler):
        raise NotImplementedError

    async def handle(self, update, context):
        raise NotImplementedError

    @staticmethod
    def _assert_can_merge(router, source):
        if type(source) is not type(router):
            raise ValueError(
                f'Mixed routers are passed to "merge" ({source} != {router})'
            )

        if source.priority is None:
            raise ValueError(
                f'Router with None priority passed to "merge" (it already was merged) ({source})'
            )

    @classmethod
    def merge(cls, source_routers):
        raise NotImplementedError


class ListRouter(Router):
    def __init__(self, priority: Optional[int] = 0):
        super().__init__(priority)
        self._handlers = []

    def add_handler(self, handler):
        self._handlers.append(handler)

    async def handle(self, update, context):
        for handler in self._handlers:
            if isinstance(handler, Router):
                handler = handler.handle

            if await handler(update, context) != SKIPPED:
                return PROCESSED

        return SKIPPED

    @classmethod
    def merge(cls, source_routers: List["ListRouter"]):
        router = cls(priority=None)

        for source in sorted(
            source_routers, reverse=True, key=lambda item: item.priority or 0
        ):
            cls._assert_can_merge(router, source)

            for handler in source._handlers:
                router.add_handler(handler)

        return router


class MapRouter(Router):
    def __init__(self, priority: Optional[int] = 0):
        super().__init__(priority)
        self._handlers: Dict[str, List] = {}

    def extract_keys(self, context):
        raise NotImplementedError

    def add_handler(self, key, handler):
        if key not in self._handlers:
            self._handlers[key] = []

        self._handlers[key].append(handler)

    async def handle(self, update, context):
        for key in self.extract_keys(context):
            for handler in self._handlers.get(key, ()):
                if isinstance(handler, Router):
                    handler = handler.handle

                if await handler(update, context) != SKIPPED:
                    return PROCESSED

        return SKIPPED

    @staticmethod
    def _merge_routers(target: "MapRouter", source: "MapRouter"):
        for key, handlers in source._handlers.items():
            for handler in handlers:
                target.add_handler(key, handler)

    @classmethod
    def merge(cls, source_routers: List["MapRouter"]):
        router = cls(priority=None)

        for source in sorted(
            source_routers, reverse=True, key=lambda item: item.priority or 0
        ):
            cls._assert_can_merge(router, source)
            cls._merge_routers(router, source)

        return router


class CommandsRouter(MapRouter):
    def add_handler(self, key, handler):
        return super().add_handler(key.lower(), handler)

    def extract_keys(self, context: Context):
        if not isinstance(context.update, Message):
            return ()

        match = re.match(
            r"\s*({prefix})?\s*({command})(.*)".format(
                prefix="|".join(
                    re.escape(prefix) for prefix in context.app.config["prefixes"]
                ),
                command="|".join(
                    re.escape(command)
                    for command in sorted(self._handlers, key=len, reverse=True)
                ),
            ),
            context.update.text,
            re.IGNORECASE,
        )

        if match is None:
            return ()

        context.prefix = match.group(1)
        context.command = match.group(2)
        context.body = (match.group(3) or "").strip()
        context.match = match

        return (context.command.lower(),)


class AttachmentsRouter(MapRouter):
    def extract_keys(self, context: Context):
        if not isinstance(context.update, Message):
            return ()

        return tuple(attachment.kind for attachment in context.update.attachments)



from typing import Dict, Optional


class OptimisticLockException(Exception):
    pass


class DocumentIsDeletedException(Exception):
    pass


class Document(dict):
    def __init__(self, *args, _storage: "Storage", _storage_key: str, **kwargs):
        self._storage = _storage
        self._storage_key = _storage_key
        super().__init__(*args, **kwargs)

    async def save(self):
        self.update(await self._storage.put(self._storage_key, self))

    async def update_and_save(self, *args, **kwargs):
        self.update(*args, **kwargs)
        await self.save()

    async def reload(self):
        data = await self._storage.get(self._storage_key)
        if data is None:
            raise DocumentIsDeletedException("Value being reloaded is deleted")
        self.clear()
        self.update(data)

    async def delete(self):
        await self._storage.delete(self._storage_key)


class Storage:
    async def init(self):
        pass

    async def put(self, key: str, data: Dict) -> Document:
        raise NotImplementedError

    async def get(self, key: str) -> Optional[Document]:
        raise NotImplementedError

    async def delete(self, key: str):
        raise NotImplementedError



# Expose public API
from .backend import Backend
from .context import Context
from .exceptions import RequestException
from .handler import PROCESSED, SKIPPED
from .helpers import get_path
from .kutana import Kutana
from .plugin import Plugin
from .storages import MemoryStorage
from .update import Attachment, Message, RecipientKind

# Attempt to enable uvloop
try:
    import uvloop

    uvloop.install()
except ModuleNotFoundError:
    pass

# Setup nice logging format
import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [ %(levelname)s ] %(message)s",
)


__all__ = [
    "Attachment",
    "Backend",
    "Context",
    "get_path",
    "Kutana",
    "MemoryStorage",
    "Message",
    "Plugin",
    "PROCESSED",
    "RecipientKind",
    "RequestException",
    "SKIPPED",
]



import functools
import re
from typing import Any, List

from .backends.vkontakte import VkontaktePluginExtension
from .handler import SKIPPED
from .router import AttachmentsRouter, CommandsRouter, ListRouter, Router
from .storage import Document
from .update import Message


class Plugin:
    """
    Plugin provides functionality to the application.

    :param name: Name of the plugin
    :param description: Description of the plugin
    """

    def __init__(self, name: str, **kwargs):
        self.app: Any

        self._hooks = []
        self._routers: List[Router] = []

        # Set some attributes to this instance
        self.name = name

        for k, v in kwargs.items():
            setattr(self, k, v)

        # Setup extensions
        self.vk = VkontaktePluginExtension(self)

    def __getattr__(self, name):
        """Defined for typing"""
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        """Defined for typing"""
        return super().__setattr__(name, value)

    def on_start(self):
        """
        Return decorator for registering coroutines that will be called
        when application starts.
        """

        def decorator(coro):
            self._hooks.append(("start", coro))
            return coro

        return decorator

    def on_completion(self):
        """
        Return decorator for registering coroutines that will be called
        after update was processed (without exception).
        """

        def decorator(coro):
            self._hooks.append(("completion", coro))
            return coro

        return decorator

    def on_exception(self):
        """
        Return decorator for registering coroutine that will be called when
        exception was raised while processing an update. It will be passed
        the update's context and raised exception as arguments.
        """

        def decorator(coro):
            self._hooks.append(("exception", coro))
            return coro

        return decorator

    def on_shutdown(self):
        """
        Return decorator for registering coroutine that will be called when
        application is being be stopped.
        """

        def decorator(coro):
            self._hooks.append(("shutdown", coro))
            return coro

        return decorator

    def on_commands(
        self,
        commands,
        priority=0,
    ):
        """
        Return decorator for registering handler that will be called
        when incoming update is a message, starts with prefix
        and one of provided commands.

        Context is automatically populated with following values:

        - prefix
        - command
        - body
        - match

        Handlers with higher priority are attempted first.

        If handler returns anything but :class:`kutana.handler.SKIPPED`,
        other handlers are not executed further.
        """

        def decorator(coro):
            router = CommandsRouter(priority=priority)
            for command in commands:
                router.add_handler(command, coro)

            self._routers.append(router)

            return coro

        return decorator

    def on_match(self, patterns, priority=0):
        """
        Return decorator for registering handler that will be called
        when incoming update is a message and it's message matches any
        specified pattern.

        Context is automatically populated with following values:

        - match

        Handlers with higher priority are attempted first.

        If handler returns anything but :class:`kutana.handler.SKIPPED`,
        other handlers are not executed further.
        """

        def decorator(func):
            router = ListRouter(priority=priority)

            @functools.wraps(func)
            async def _wrapper(update, ctx):
                if not isinstance(update, Message):
                    return SKIPPED

                for pattern in patterns:
                    match = re.match(pattern, update.text)
                    if match:
                        break
                else:
                    return SKIPPED

                ctx.match = match

                return await func(update, ctx)

            router.add_handler(_wrapper)

            self._routers.append(router)

            return func

        return decorator

    def on_attachments(
        self,
        kinds,
        priority=0,
    ):
        """
        Return decorator for registering handler that will be called
        when incoming update is a message and have at least one
        attachment with one of specified types.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and return values.
        """

        def decorator(coro):
            router = AttachmentsRouter(priority=priority)
            for kind in kinds:
                router.add_handler(kind, coro)

            self._routers.append(router)

            return coro

        return decorator

    def on_messages(self, priority=-1):
        """
        Return decorator for registering handler that will be called
        when incoming update is a message. Handler will always be
        called unless any other handler with lower priority returned
        appropriate value.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and return values.
        """

        def decorator(coro):
            router = ListRouter(priority=priority)

            @functools.wraps(coro)
            async def _wrapper(update, context):
                if not isinstance(update, Message):
                    return SKIPPED
                return await coro(update, context)

            router.add_handler(_wrapper)

            self._routers.append(router)

            return coro

        return decorator

    def on_updates(self, priority=0):
        """
        Return decorator for registering handler that will be always
        called (for messages and not messages).

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and return values.
        """

        def decorator(coro):
            router = ListRouter(priority=priority)
            router.add_handler(coro)

            self._routers.append(router)

            return coro

        return decorator

    def with_storage(self, check_sender=None, check_recipient=None, storage="default"):
        """
        This decorator allow plugins to implicitly require access to database.
        Context is populated with the following fields:

        - "storage" - Storage object that can be used to get, update or delete
            data, stored with specific keys.
        - "sender" - document with data, previously saved to storage for
            sender of the message.
        - "recipient" - document with data, previously saved to storage for
            recipient of the message.

        This decorator also accepts arguments `check_sender` and `check_recipient`.
        You can specify dictionaries that treated as expected values in corresponding
        data objects. You can use it like `with_storage(check_sender={"state": None})`.
        You can use specific storage by providing "storage" argument.

        Dcoument is an dictionary with additional async methods:

        - "save" - save document's data to database.
        - "update_and_save" - same as "update" and then "save".
        - "reload" - reload data from database.
        - "delete" - deletes data from database.

        Any attempt to save data to database can raise OptimisticLockException, that
        means that your code was working at least partially with outdated data. You
        can either ignore it or retry your operation after reloading data.

        Example:

        .. code-block:: python

            context.sender.update({"field1": "value1"})  # not yet stored in database
            await context.sender.save()  # saves data in database

            context.sender.update_and_save(field2="value2")  # update data and store it in database
        """

        def _perform_check(data, check):
            """Return true if handler should be called."""

            return not check or all(data.get(k) == v for k, v in check.items())

        def decorator(coro):
            @functools.wraps(coro)
            async def wrapper(update, context):
                context.storage = self.app.storages[storage]

                if not getattr(context, "sender", None):
                    context.sender = await context.storage.get(context.sender_unique_id)
                    if not context.sender:
                        context.sender = Document(
                            _storage=context.storage,
                            _storage_key=context.sender_unique_id,
                        )

                if not _perform_check(context.sender, check_sender):
                    return SKIPPED

                if not getattr(context, "recipient", None):
                    context.recipient = await context.storage.get(
                        context.recipient_unique_id
                    )
                    if not context.recipient:
                        context.recipient = Document(
                            _storage=context.storage,
                            _storage_key=context.recipient_unique_id,
                        )

                if not _perform_check(context.recipient, check_recipient):
                    return SKIPPED

                return await coro(update, context)

            return wrapper

        return decorator



from typing import Union

from .helpers import chunks
from .update import Message


class Context:
    def __init__(self, app, update, backend):
        self.app = app
        self.update: Union[Message, dict] = update
        self.backend = backend

    def __getattr__(self, name):
        """Defined for typing"""
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        """Defined for typing"""
        return super().__setattr__(name, value)

    @property
    def sender_unique_id(self):
        if isinstance(self.update, Message):
            sender_id = self.update.sender_id
        elif hasattr(self, "sender_id"):
            sender_id = self.sender_id
        else:
            raise ValueError("Can't form an unique sender id")

        return f"{self.backend.get_identity()}:s:{sender_id}"

    @property
    def recipient_unique_id(self):
        if isinstance(self.update, Message):
            recipient_id = self.update.recipient_id
        elif hasattr(self, "recipient_id"):
            recipient_id = self.recipient_id
        else:
            raise ValueError("Can't form an unique recipient id")

        return f"{self.backend.get_identity()}:r:{recipient_id}"

    @property
    def request(self):
        return self.backend.request

    async def send_message(self, recipient_id, text=None, attachments=None, **kwargs):
        results = []

        text_segments = list(chunks(str(text))) if text else [None]

        for text_segment in text_segments[:-1]:
            results.append(await self.backend.send_message(recipient_id, text_segment))

        results.append(
            await self.backend.send_message(
                recipient_id,
                text_segments[-1],
                attachments,
                **kwargs,
            )
        )

        return results

    async def reply(self, text=None, attachments=None, **kwargs):
        if isinstance(self.update, Message):
            recipient_id = self.update.recipient_id
        elif hasattr(self, "recipient_id"):
            recipient_id = self.recipient_id
        else:
            raise ValueError("You can't reply to this update")

        return await self.send_message(
            recipient_id,
            text,
            attachments,
            **kwargs,
        )


def split_large_text(text, length=4096):
    """
    Split text into chunks with specified length.

    :param text: text for splitting
    :param length: maximum size for one chunk
    :returns: tuple of chunks
    """

    text = str(text)

    yield text[0:length]

    for i in range(length, len(text), length):
        yield text[i : i + length]



import argparse
import logging
import os
from importlib import import_module

import yaml

from . import Kutana
from .backends import Telegram, VkontakteCallback, VkontakteLongpoll
from .loaders import load_plugins_from_module, load_plugins_from_path
from .storages import MemoryStorage, MongoDBStorage, SqliteStorage


class CuteSafeDumper(yaml.SafeDumper):
    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)


def get_parser():
    parser = argparse.ArgumentParser("kutana", description="helpfull cli utility")
    subparsers = parser.add_subparsers(dest="command")

    # Project initialization
    parser_init = subparsers.add_parser("init", help="initiate kutana project")
    parser_init.add_argument(
        "path",
        type=str,
        default=None,
        help="path to folder to populate with default project config",
    )

    # Project running
    parser_run = subparsers.add_parser(
        "run",
        help="run kutana project using provided config (working directory will be changed to the one with config file)",
    )
    parser_run.add_argument(
        "path",
        type=str,
        default="config.yml",
        help="path to config of the project",
    )
    parser_run.add_argument(
        "--debug",
        dest="debug",
        default=False,
        action="store_true",
        help="set logging level to debug",
    )

    return parser


def initiate_project(args):
    project_root = args.path or "."

    if os.path.isfile(project_root):
        logging.error(f"Specified project folder is a plain file: {project_root}")
        exit(1)

    if not os.path.isdir(project_root):
        os.makedirs(project_root)

    config_file = os.path.join(project_root, "config.yml")

    if os.path.isfile(config_file):
        logging.error(
            f"Specified project folder already has configuration: {project_root}/config.yml"
        )
        exit(1)

    with open(config_file, "w") as fh:
        yaml.dump(
            {
                "backends": [],
                "plugins": ["plugins/"],
                "prefixes": [".", "!", "/"],
                "storages": [],
            },
            fh,
            Dumper=CuteSafeDumper,
        )


def run_project(args):  # noqa:C901
    # Load configuration
    with open(args.path) as fh:
        config = yaml.safe_load(fh)

    # Change the current working directory to one where config is if needed
    target_directory = os.path.dirname(args.path)
    if target_directory:
        os.chdir(target_directory)

    # Change logging level if needed
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    # Initiate application
    app = Kutana()

    # Populate configuration
    app.config.update(config)

    # Add storages
    for storage_config in config.get("storages", ()):
        storage_name = storage_config["name"]

        if storage_config["kind"] == "memory":
            storage_class = MemoryStorage
        elif storage_config["kind"] == "mongodb":
            storage_class = MongoDBStorage
        elif storage_config["kind"] == "sqlite":
            storage_class = SqliteStorage
        else:
            raise ValueError(f'Unknown storage kind: "{storage_config["kind"]}"')

        storage_kwargs = {**storage_config}
        storage_kwargs.pop("name")
        storage_kwargs.pop("kind")

        app.add_storage(storage_name, storage_class(**storage_kwargs))

    # Add backends
    for backend_config in config.get("backends", ()):
        if backend_config["kind"] == "vk-lp":
            backend_class = VkontakteLongpoll
        elif backend_config["kind"] == "vk-cb":
            backend_class = VkontakteCallback
        elif backend_config["kind"] == "tg":
            backend_class = Telegram
        else:
            raise ValueError(f'Unknown backend kind: "{backend_config["kind"]}"')

        backend_kwargs = {**backend_config}
        backend_kwargs.pop("kind")

        app.add_backend(backend_class(**backend_kwargs))

    # Add plugins
    for plugin_config in config.get("plugins", ()):
        if isinstance(plugin_config, str):
            for plugin in load_plugins_from_path(plugin_config):
                app.add_plugin(plugin)
        elif not isinstance(plugin_config, dict) or len(plugin_config) != 1:
            raise ValueError(f'Unknown plugin config format: "{plugin_config}"')
        elif plugin_config.get("path"):
            for plugin in load_plugins_from_path(plugin_config["path"]):
                app.add_plugin(plugin)
        elif plugin_config.get("module"):
            app.add_plugin(
                load_plugins_from_module(import_module(plugin_config["module"]))
            )
        else:
            raise ValueError(f'Unknown plugin config kind: "{plugin_config}"')

    # Run the application
    app.run()


if __name__ == "__main__":
    args = get_parser().parse_args()

    if args.command == "init":
        initiate_project(args)
    elif args.command == "run":
        run_project(args)



class RequestException(Exception):
    def __init__(self, backend, method, kwargs, response, exception=None):
        self.backend = backend
        self.method = method
        self.kwargs = kwargs
        self.response = response
        self.exception = exception



import json
import sqlite3
import threading
from contextlib import contextmanager

from ..storage import Document, OptimisticLockException, Storage


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class SqliteStorage(Storage):
    """
    Storage implementation of the storage that uses sqlite3.
    """

    def __init__(self, path):
        self._path = path
        self._lock = threading.Lock()
        self.connection: sqlite3.Connection

    async def init(self):
        self.connection = sqlite3.connect(self._path, check_same_thread=False)
        self.connection.row_factory = dict_factory

        with self.cursor() as cur:
            cur.execute(
                """
                CREATE TABLE IF NOT EXISTS kvs (
                    key text NOT NULL,
                    val text NOT NULL,
                    ver integer NOT NULL default 0
                )
            """
            )
            cur.execute("CREATE UNIQUE INDEX IF NOT EXISTS kvs__key ON kvs (key)")
            cur.execute("CREATE INDEX IF NOT EXISTS kvs__key__ver ON kvs (key, ver)")

    @contextmanager
    def cursor(self):
        with self.connection:
            yield self.connection.cursor()

    async def put(self, key, data):
        with self._lock:
            old_version = data.get("_version") or 0
            new_version = old_version + 1
            new_data = {k: v for k, v in data.items() if v is not None}
            serialized_data = json.dumps(new_data, ensure_ascii=False)

            try:
                with self.cursor() as cur:
                    if old_version:
                        cur.execute(
                            "UPDATE kvs SET val = ?, ver = ? WHERE key = ? AND ver = ?",
                            (serialized_data, new_version, key, old_version),
                        )

                    if not old_version or cur.rowcount < 1:
                        cur.execute(
                            "INSERT INTO kvs (key, val, ver) VALUES (?, ?, ?)",
                            (key, serialized_data, new_version),
                        )
            except sqlite3.IntegrityError:
                raise OptimisticLockException(
                    f"Failed to update data for key {key} (mismatched version)"
                )

            return Document(
                {**new_data, "_version": new_version}, _storage=self, _storage_key=key
            )

    async def get(self, key):
        with self._lock:
            with self.cursor() as cur:
                cur.execute("SELECT * FROM kvs WHERE key = ?", (key,))
                row = cur.fetchone()

                if not row:
                    return None

                return Document(
                    {**json.loads(row["val"]), "_version": row["ver"]},
                    _storage=self,
                    _storage_key=key,
                )

    async def delete(self, key):
        with self._lock:
            with self.cursor() as cur:
                cur.execute("DELETE FROM kvs WHERE key = ?", (key,))



from .callback import VkontakteCallback
from .extensions import VkontaktePluginExtension
from .longpoll import VkontakteLongpoll

__all__ = [
    "VkontakteCallback",
    "VkontaktePluginExtension",
    "VkontakteLongpoll",
]



import pymongo
import pymongo.errors
from motor.motor_asyncio import (
    AsyncIOMotorClient,
    AsyncIOMotorCollection,
    AsyncIOMotorDatabase,
)

from ..storage import Document, OptimisticLockException, Storage


class MongoDBStorage(Storage):
    """
    Storage implementation of the storage that uses a running MongoDB server.
    """

    def __init__(self, mongodb_uri, database="kutana", collection="storage"):
        self._config = {
            "host": mongodb_uri,
            "database": database,
            "collection": collection,
        }
        self.client: AsyncIOMotorClient
        self.database: AsyncIOMotorDatabase
        self.collection: AsyncIOMotorCollection

    async def init(self):
        self.client = AsyncIOMotorClient(self._config["host"])
        self.database = self.client[self._config["database"]]
        self.collection = self.database[self._config["collection"]]

    async def put(self, key, data):
        try:
            data = await self.collection.find_one_and_update(
                {
                    "_id": key,
                    "_version": data.get("_version"),
                },
                {
                    "$set": {
                        **{k: v for k, v in data.items() if v is not None},
                        "_id": key,
                        "_version": data.get("_version", 0) + 1,
                    },
                    "$unset": {
                        **{k: v for k, v in data.items() if v is None},
                    },
                },
                upsert=True,
                return_document=pymongo.ReturnDocument.AFTER,
            )

            data.pop("_id")  # Key should not be in resulting data

            return Document(data, _storage=self, _storage_key=key)
        except pymongo.errors.DuplicateKeyError:  # type: ignore
            raise OptimisticLockException(
                f'Failed to update data for key "{key}" (mismatched version)'
            )

    async def get(self, key):
        data = await self.collection.find_one({"_id": key}, projection={"_id": 0})

        if data:
            return Document(data, _storage=self, _storage_key=key)

        return data

    async def delete(self, key):
        await self.collection.delete_one({"_id": key})



from functools import wraps

from .backend import Backend
from .context import Context
from .handler import SKIPPED
from .update import Message, RecipientKind


def expect_recipient_kind(expected_recipient_kind: RecipientKind):
    """
    Return decorators that skips all messages with "recipient_kind"
    different from specified one.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(message: Message, context: Context):
            if message.recipient_kind != expected_recipient_kind:
                return SKIPPED
            return await func(message, context)

        return wrapper

    return decorator


def expect_backend(expected_identity):
    """
    Return decorators that skips all updates acquired from
    backends with identity different from specified one.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(message: Message, context: Context):
            if context.backend.get_identity() != expected_identity:
                return SKIPPED
            return await func(message, context)

        return wrapper

    return decorator


async def _get_user_statuses_vk(backend, chat_id, user_id):
    members_response = await backend.request(
        "messages.getConversationMembers",
        {"peer_id": chat_id, "fields": ""},
    )

    for member in members_response["items"]:
        if member["member_id"] == user_id:
            statuses = []

            if member.get("is_owner"):
                statuses.append("admin")
                statuses.append("owner")
            elif member.get("is_admin"):
                statuses.append("admin")

            return statuses

    return []


async def _get_user_statuses_tg(backend, chat_id, user_id):
    chat_administrators = await backend.request(
        "getChatAdministrators",
        {"chat_id": chat_id},
    )

    for administrator in chat_administrators:
        if administrator["user"]["id"] == user_id:
            statuses = []

            if administrator["status"] == "creator":
                statuses.extend(["admin", "owner"])
            elif administrator["status"] == "administrator":
                statuses.append("admin")

            return statuses

    return []


async def _get_user_statuses(backend: Backend, chat_id, user_id):
    if backend.get_identity() == "vk":
        return await _get_user_statuses_vk(backend, chat_id, user_id)

    if backend.get_identity() == "tg":
        return await _get_user_statuses_tg(backend, chat_id, user_id)

    raise NotImplementedError()


def _expect_sender_status(func, expected_status):
    @expect_recipient_kind(RecipientKind.GROUP_CHAT)
    @wraps(func)
    async def wrapper(message: Message, context: Context):
        sender_statuses = await _get_user_statuses(
            context.backend,
            message.recipient_id,
            message.sender_id,
        )

        if expected_status not in sender_statuses:
            return SKIPPED

        return await func(message, context)

    return wrapper


def expect_sender_is_admin():
    """
    Return decorators that skips all messages that were
    send by non-admin users (we treat owners as admins).
    """

    def decorator(func):
        return _expect_sender_status(func, "admin")

    return decorator


def expect_sender_is_owner():
    """
    Return decorators that skips all messages that were
    send by non-owner users.
    """

    def decorator(func):
        return _expect_sender_status(func, "owner")

    return decorator



import json

from ...helpers import pick, uniq_by
from ...router import MapRouter
from ...update import Message


class VkontakteChatActionRouter(MapRouter):
    def extract_keys(self, context):
        if not isinstance(context.update, Message):
            return ()

        if context.backend.get_identity() != "vk":
            return ()

        chat_action = context.update.raw["object"]["message"].get("action")
        if not chat_action:
            return ()

        context.chat_action = chat_action

        return (chat_action["type"],)


class VkontaktePayloadRouter(MapRouter):
    def __init__(self, priority=0):
        super().__init__(priority)
        self._possible_key_shapes = set()

    def _update_possible_key_shapes(self, key):
        if not isinstance(key, dict):
            return

        self._possible_key_shapes = uniq_by(
            [
                *self._possible_key_shapes,
                tuple(sorted(key.keys())),
            ]
        )

    def add_handler(self, key, handler):
        self._update_possible_key_shapes(key)
        return super().add_handler(self._to_hashable(key), handler)

    @staticmethod
    def _merge_routers(target, source):
        super()._merge_routers(target, source)
        target._possible_key_shapes.update(source._possible_key_shapes)

    def _to_hashable(self, value):
        if isinstance(value, dict):
            return tuple((k, self._to_hashable(v)) for k, v in sorted(value.items()))

        if isinstance(value, list):
            return tuple(self._to_hashable(item) for item in value)

        return value

    def extract_keys(self, context):
        if not isinstance(context.update, Message):
            return ()

        if context.backend.get_identity() != "vk":
            return ()

        raw_payload = context.update.raw["object"]["message"].get("payload")
        if not raw_payload:
            return ()

        try:
            payload = json.loads(raw_payload)
        except json.JSONDecodeError:
            return ()

        context.payload = payload

        if isinstance(payload, dict):
            return (
                self._to_hashable(pick(payload, possible_key_shape))
                for possible_key_shape in self._possible_key_shapes
            )
        else:
            return (self._to_hashable(payload),)


class VkontakteCallbackRouter(VkontaktePayloadRouter):
    def extract_keys(self, context):
        if isinstance(context.update, Message):
            return ()

        if context.backend.get_identity() != "vk":
            return ()

        if context.update["type"] != "message_event":
            return ()

        payload = context.update["object"]["payload"]

        context.payload = payload

        async def _send_message_event_answer(event_data, **kwargs):
            return await context.request(
                "messages.sendMessageEventAnswer",
                {
                    "event_id": context.update["object"]["event_id"],
                    "user_id": context.update["object"]["user_id"],
                    "peer_id": context.update["object"]["peer_id"],
                    "event_data": json.dumps(event_data),
                    **kwargs,
                },
            )

        context.send_message_event_answer = _send_message_event_answer

        if isinstance(payload, dict):
            return (
                self._to_hashable(pick(payload, possible_key_shape))
                for possible_key_shape in self._possible_key_shapes
            )
        else:
            return (self._to_hashable(payload),)


class VkontaktePluginExtension:
    def __init__(self, plugin):
        self._plugin = plugin

    def on_chat_actions(
        self,
        kinds,
        priority=0,
    ):
        """
        Return decorator for registering handler that will be called if
        incoming update is a message with action (only for chats).

        Context is automatically populated with following values:

        - "action" - chat action object.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and return values.
        """

        def decorator(coro):
            chat_action_router = VkontakteChatActionRouter(priority=priority)
            for kind in kinds:
                chat_action_router.add_handler(kind, coro)

            self._plugin._routers.append(chat_action_router)

            return coro

        return decorator

    def on_payloads(
        self,
        payloads,
        priority=0,
    ):
        """
        Return decorator for registering handler that will be called if
        incoming update is a message with specified value in payload.
        Payload is treated as JSON-encoded value. Excessive fields in
        payload are ignored. Use strings or lists for exact matching.

        Context is automatically populated with following values:

        - "payload" - chat action object.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and return values.
        """

        def decorator(coro):
            payload_router = VkontaktePayloadRouter(priority=priority)
            for payload in payloads:
                payload_router.add_handler(payload, coro)

            self._plugin._routers.append(payload_router)

            return coro

        return decorator

    def on_callbacks(
        self,
        payloads,
        priority=0,
    ):
        """
        Return decorator for registering handler that will be called if
        incoming update is a message_event with specified value in payload.
        Excessive fields in payload are ignored. Use strings or lists for
        exact matching.

        Context is automatically populated with following values:

        - "payload" - callback payload value.
        - "send_message_event_answer" - helper method for "messages.sendMessageEventAnswer".

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and return values.
        """

        def decorator(coro):
            callback_router = VkontakteCallbackRouter(priority=priority)
            for payload in payloads:
                callback_router.add_handler(payload, coro)

            self._plugin._routers.append(callback_router)

            return coro

        return decorator



import asyncio
import json
import logging
import re
from functools import partial
from itertools import zip_longest
from random import random

from kutana.helpers import create_httpx_async_client

from ...backend import Backend
from ...exceptions import RequestException
from ...update import Attachment, AttachmentKind, Message, RecipientKind

DEFAULT_RECEIVABLE_EVENTS = {
    "message_new": 1,
    "message_reply": 0,
    "message_edit": 1,
    "message_allow": 0,
    "message_deny": 0,
    "photo_new": 0,
    "audio_new": 0,
    "video_new": 0,
    "wall_reply_new": 0,
    "wall_reply_edit": 0,
    "wall_reply_delete": 0,
    "wall_post_new": 0,
    "wall_repost": 0,
    "board_post_new": 0,
    "board_post_edit": 0,
    "board_post_delete": 0,
    "board_post_restore": 0,
    "photo_comment_new": 0,
    "photo_comment_edit": 0,
    "photo_comment_delete": 0,
    "photo_comment_restore": 0,
    "video_comment_new": 0,
    "video_comment_edit": 0,
    "video_comment_delete": 0,
    "video_comment_restore": 0,
    "market_comment_new": 0,
    "market_comment_edit": 0,
    "market_comment_delete": 0,
    "market_comment_restore": 0,
    "poll_vote_new": 0,
    "group_join": 0,
    "group_leave": 0,
    "user_block": 0,
    "user_unblock": 0,
    "group_change_settings": 1,
    "group_change_photo": 0,
    "group_officers_edit": 0,
    "donut_subscription_create": 0,
    "donut_subscription_prolonged": 0,
    "donut_subscription_expired": 0,
    "donut_subscription_cancelled": 0,
    "subscription_price_changed": 0,
    "donut_money_withdraw": 0,
    "donut_money_withdraw_error": 0,
}


class Vkontakte(Backend):
    def __init__(
        self,
        token,
        requests_per_second=19,
        api_version="5.131",
        api_url="https://api.vk.com",
    ):
        if not token:
            raise ValueError("No token provided for backend")

        self.api_token = token
        self.api_version = api_version
        self.api_request_pause = 1 / requests_per_second

        self.client = create_httpx_async_client()

        self.group: dict
        self.requests_queue: asyncio.Queue
        self.requests_queue_handler: asyncio.Task

        self.api_request_url = (
            api_url + f"/method/{{}}?access_token={token}&v={api_version}"
        )

    def get_identity(self):
        return "vk"

    async def _resolve_screen_name(self, screen_name, cache={}):
        if cache and screen_name in cache:
            return cache[screen_name]

        response = await self._direct_request(
            "utils.resolveScreenName", {"screen_name": screen_name}
        )

        if response and isinstance(cache, dict):
            if len(cache) > 65536:
                cache.clear()
            cache[screen_name] = response

        return response

    async def setup_context(self, context):
        context.resolve_screen_name = self._resolve_screen_name

    def _extract_text(self, text):
        bot_mention_regex = r"\[((public|club){}|{})\|.+\]".format(
            re.escape(str(self.group["id"])),
            re.escape(self.group["screen_name"]),
        )

        return re.sub(bot_mention_regex, "", text).lstrip()

    async def _get_file(self, url):
        response = await self.client.get(url)
        return response.content

    @staticmethod
    def _make_attachment_full_id(raw_attachment_type, raw_attachment_media):
        full_id_postfix = "_".join(
            filter(
                None,
                map(
                    str,
                    [
                        raw_attachment_media.get("owner_id"),
                        raw_attachment_media.get("id"),
                        raw_attachment_media.get("access_key"),
                    ],
                ),
            )
        )

        if not full_id_postfix:
            return None

        return f"{raw_attachment_type}{full_id_postfix}"

    def _make_attachment(self, raw_attachment):
        raw_attachment_media = raw_attachment[raw_attachment["type"]]

        full_id = self._make_attachment_full_id(
            raw_attachment["type"], raw_attachment_media
        )

        if raw_attachment["type"] == "audio":
            return Attachment(
                id=full_id,
                kind=AttachmentKind.AUDIO,
                title=f"{raw_attachment_media['artist']} - {raw_attachment_media['title']}",
                raw=raw_attachment,
                get_file=partial(self._get_file, raw_attachment_media["url"]),
            )

        if raw_attachment["type"] == "doc":
            return Attachment(
                id=full_id,
                kind=AttachmentKind.DOCUMENT,
                title=raw_attachment_media["title"],
                raw=raw_attachment,
                get_file=partial(self._get_file, raw_attachment_media["url"]),
            )

        if raw_attachment["type"] == "photo":
            largest_size = list(
                sorted(raw_attachment_media["sizes"], key=lambda size: size["width"])
            )[0]
            return Attachment(
                id=full_id,
                kind=AttachmentKind.IMAGE,
                raw=raw_attachment,
                get_file=partial(self._get_file, largest_size["url"]),
            )

        if raw_attachment["type"] == "video":
            return Attachment(
                id=full_id,
                kind=AttachmentKind.VIDEO,
                raw=raw_attachment,
                get_file=None,
            )

        if raw_attachment["type"] == "audio_message":
            return Attachment(
                id=full_id,
                kind=AttachmentKind.VOICE,
                raw=raw_attachment,
                get_file=partial(self._get_file, raw_attachment_media["link_mp3"]),
            )

        return Attachment(
            id=full_id,
            kind=raw_attachment["type"],
            raw=raw_attachment,
        )

    def _make_update(self, raw_update):
        if raw_update["type"] != "message_new":
            return raw_update

        message = raw_update["object"]["message"]

        attachments = []

        for attachment in message["attachments"]:
            attachments.append(self._make_attachment(attachment))

        if message["peer_id"] > 2000000000:
            recipient_kind = RecipientKind.GROUP_CHAT
        else:
            recipient_kind = RecipientKind.PRIVATE_CHAT

        return Message(
            sender_id=message["from_id"],
            recipient_id=message["peer_id"],
            recipient_kind=recipient_kind,
            text=self._extract_text(raw_update["object"]["message"]["text"]),
            attachments=attachments,
            date=message["date"],
            raw=raw_update,
        )

    async def _update_group_data(self):
        groups = await self._direct_request("groups.getById", {"fields": "screen_name"})

        self.group = groups[0]

    async def on_start(self, app):
        await self._update_group_data()

        logging.info(
            'logged in as "%s" ( https://vk.com/club%s )',
            self.group["name"],
            self.group["id"],
        )

        self.requests_queue = asyncio.Queue()

        self.requests_queue_handler = asyncio.ensure_future(
            self._handle_requests_queue()
        )

    async def _direct_request(self, method, kwargs):
        response = await self.client.post(
            self.api_request_url.format(method),
            data={k: v for k, v in kwargs.items() if v is not None},
        )

        try:
            data = response.json()
        except Exception as exception:
            raise RequestException(
                self,
                method,
                kwargs,
                response.content,
                exception=exception,
            )

        if data.get("response") is None or data.get("execute_errors"):
            raise RequestException(self, method, kwargs, data)

        return data["response"]

    async def _handle_requests_queue(self):
        while True:
            await asyncio.sleep(self.api_request_pause)

            requests_chunk = []

            for _ in range(25):
                try:
                    requests_chunk.append(self.requests_queue.get_nowait())
                except asyncio.QueueEmpty:
                    break

            if not requests_chunk:
                continue

            asyncio.ensure_future(self._handle_requests_chunk(requests_chunk))

    async def _handle_requests_chunk(self, chunk):
        code = "return ["

        for method, kwargs, _ in chunk:
            serialized_kwargs = json.dumps(
                kwargs, separators=(",", ":"), ensure_ascii=False
            )
            code += f"API.{method}({serialized_kwargs}),"

        code += "];"

        response = await self._direct_request("execute", {"code": code})

        for result, (method, kwargs, future) in zip_longest(response, chunk):
            if future.done():
                continue

            if result is not None:
                future.set_result(result)
                continue

            future.set_exception(
                RequestException(
                    self,
                    method,
                    kwargs,
                    result,
                ),
            )

    async def request(self, method, kwargs):
        future = asyncio.Future()
        await self.requests_queue.put((method, kwargs, future))
        return await future

    async def upload_attachment(self, attachment, peer_id):
        if attachment.kind == AttachmentKind.IMAGE:
            upload_data = await self.request(
                "photos.getMessagesUploadServer",
                {"peer_id": peer_id},
            )

            upload_response = await self.client.post(
                upload_data["upload_url"],
                files={"photo": attachment.content},
            )

            result = await self.request(
                "photos.saveMessagesPhoto",
                upload_response.json(),
            )

            return f'photo{result[0]["owner_id"]}_{result[0]["id"]}'

        if attachment.kind in (AttachmentKind.DOCUMENT, AttachmentKind.VOICE):
            vk_type = (
                "audio_message" if attachment.kind == AttachmentKind.VOICE else "doc"
            )

            upload_data = await self.request(
                "docs.getMessagesUploadServer",
                {"peer_id": peer_id, "type": vk_type},
            )

            upload_response = await self.client.post(
                upload_data["upload_url"],
                files={"file": attachment.content},
            )

            result = await self.request(
                "docs.save",
                upload_response.json(),
            )

            return f'{vk_type}{result[vk_type]["owner_id"]}_{result[vk_type]["id"]}'

        raise ValueError(f'Unsupported attachment kind: "{attachment.kind}"')

    async def send_message(self, recipient_id, text=None, attachments=None, **kwargs):
        formatted_attachments = []

        if attachments:
            for attachment in attachments:
                if isinstance(attachment, str):
                    formatted_attachments.append(attachment)
                    continue

                if attachment.id:
                    formatted_attachments.append(attachment.id)
                    continue

                formatted_attachments.append(
                    await self.upload_attachment(attachment, peer_id=recipient_id),
                )

        return await self.request(
            "messages.send",
            {
                "peer_id": recipient_id,
                "message": str(text),
                "attachment": ",".join(formatted_attachments),
                **kwargs,
                "random_id": kwargs.get("random_id")
                or int(random() * 4294967296) - 2147483648,
            },
        )

    async def on_shutdown(self, app):
        self.requests_queue_handler.cancel()
        await self.client.aclose()



import asyncio
import io
import json
import logging
from functools import partial

from ..backend import Backend
from ..exceptions import RequestException
from ..helpers import create_httpx_async_client, pick_by
from ..update import Attachment, AttachmentKind, Message, RecipientKind

SUPPORTED_ATTACHMENT_KINDS = {
    AttachmentKind.AUDIO: ("sendAudio", "audio"),
    AttachmentKind.DOCUMENT: ("sendDocument", "document"),
    AttachmentKind.IMAGE: ("sendPhoto", "photo"),
    AttachmentKind.VIDEO: ("sendVideo", "video"),
    AttachmentKind.VOICE: ("sendVoice", "voice"),
}

TELEGRAM_POSSIBLE_ATTACHMENT_KINDS = [
    "audio",
    "contact",
    "document",
    "invoice",
    "location",
    "passport_data",
    "photo",
    "poll",
    "sticker",
    "video_note",
    "video",
    "voice",
]


class Telegram(Backend):
    def __init__(
        self,
        token,
        messages_per_second=29,
        api_url="https://api.telegram.org",
        **kwargs,
    ):
        super().__init__(**kwargs)

        if not token:
            raise ValueError("No token provided for backend")

        self.client = create_httpx_async_client()

        self.username = None
        self.api_token = token
        self.api_messages_pause = 1 / messages_per_second
        self.api_messages_lock: asyncio.Lock

        api_url = api_url.rstrip("/")
        self.api_url = f"{api_url}/bot{token}/{{}}"
        self.file_url = f"{api_url}/file/bot{token}/{{}}"

    def get_identity(self):
        return "tg"

    async def setup_context(self, context):
        if isinstance(context.update, dict):
            if context.update.get("callback_query"):
                context.sender_id = context.update["callback_query"]["from"]["id"]
                context.recipient_id = context.update["callback_query"]["message"][
                    "chat"
                ]["id"]

    async def on_start(self, app):
        me = await self.request("getMe", {})

        name = me.get("first_name", "") + " " + me.get("last_name", "")
        name = name.strip() or "(unknown)"

        logging.info('logged in as "%s" ( https://t.me/%s )', name, me["username"])

        self.username = me["username"]

        self.api_messages_lock = asyncio.Lock()

    def _is_file_like_value(self, value):
        if isinstance(value, (io.IOBase, bytes)):
            return True

        if (
            isinstance(value, tuple)
            and len(value) == 2
            and isinstance(value[1], (io.IOBase, bytes))
        ):
            return True

        return False

    async def request(self, method, kwargs):
        data = {}
        files = {}

        for key, value in kwargs.items():
            if value is None:
                continue

            if self._is_file_like_value(value):
                files[key] = value
            elif isinstance(value, (list, dict)):
                data[key] = json.dumps(value)
            else:
                data[key] = value

        response = await self.client.post(
            self.api_url.format(method),
            data=data,
            files=files,
            timeout=60,
        )

        response_data = response.json()

        if not response_data.get("ok"):
            raise RequestException(self, method, kwargs, response_data)

        return response_data["result"]

    async def _get_file(self, file_id):
        file = await self.request("getFile", {"file_id": file_id})
        response = await self.client.get(self.file_url.format(file["file_path"]))
        return response.content

    def _make_attachment(self, raw_attachment, raw_attachment_kind):
        if raw_attachment_kind == "audio":
            return Attachment(
                id=raw_attachment["file_id"],
                kind=AttachmentKind.AUDIO,
                title=raw_attachment["title"],
                raw=raw_attachment,
                get_file=partial(self._get_file, raw_attachment["file_id"]),
            )

        if raw_attachment_kind == "document":
            return Attachment(
                id=raw_attachment["file_id"],
                kind=AttachmentKind.DOCUMENT,
                title=raw_attachment.get("file_name"),
                raw=raw_attachment,
                get_file=partial(self._get_file, raw_attachment["file_id"]),
            )

        if raw_attachment_kind == "photo":
            largest_photo = list(
                sorted(raw_attachment, key=lambda photo: photo["width"])
            )[0]
            return Attachment(
                id=largest_photo["file_id"],
                kind=AttachmentKind.IMAGE,
                raw=raw_attachment,
                get_file=partial(self._get_file, largest_photo["file_id"]),
            )

        if raw_attachment_kind == "video":
            return Attachment(
                id=raw_attachment["file_id"],
                kind=AttachmentKind.VIDEO,
                title=raw_attachment.get("file_name"),
                raw=raw_attachment,
                get_file=partial(self._get_file, raw_attachment["file_id"]),
            )

        if raw_attachment_kind == "voice":
            return Attachment(
                id=raw_attachment["file_id"],
                kind=AttachmentKind.VOICE,
                raw=raw_attachment,
                get_file=partial(self._get_file, raw_attachment["file_id"]),
            )

        return Attachment(
            id=raw_attachment.get("file_id"),
            kind=raw_attachment_kind,
            raw=raw_attachment,
        )

    def _extract_text(self, update):
        if not update["message"].get("entities"):
            return update["message"].get("text")

        original_text = update["message"]["text"]
        text = ""
        processed_index = 0

        for entity in update["message"]["entities"]:
            if entity["type"] != "bot_command":
                continue

            text += original_text[processed_index : entity["offset"]]
            new_processed_index = entity["offset"] + entity["length"]
            command = original_text[entity["offset"] : new_processed_index]
            processed_index = new_processed_index

            if command.endswith(f"@{self.username}"):
                text += command[: -len(f"@{self.username}")]
            else:
                text += command

        return text + original_text[processed_index:]

    def _make_update(self, raw_update):
        if "message" not in raw_update:
            return raw_update

        attachments = []

        for kind in TELEGRAM_POSSIBLE_ATTACHMENT_KINDS:
            if raw_update["message"].get(kind):
                attachments.append(
                    self._make_attachment(raw_update["message"][kind], kind)
                )

        if raw_update["message"]["chat"]["type"] == "private":
            recipient_kind = RecipientKind.PRIVATE_CHAT
        else:
            recipient_kind = RecipientKind.GROUP_CHAT

        return Message(
            sender_id=raw_update["message"]["from"]["id"],
            recipient_id=raw_update["message"]["chat"]["id"],
            recipient_kind=recipient_kind,
            text=self._extract_text(raw_update) or "",
            attachments=attachments,
            date=raw_update["message"]["date"],
            raw=raw_update,
        )

    async def acquire_updates(self, queue: asyncio.Queue):
        offset = 0

        while True:
            try:
                response = await self.request(
                    "getUpdates", {"timeout": 25, "offset": offset}
                )
            except asyncio.CancelledError:
                raise
            except Exception:
                logging.exception("Exceptions while gettings updates (Telegram)")
                await asyncio.sleep(1)
                continue

            for raw_update in response:
                await queue.put((self._make_update(raw_update), self))
                offset = raw_update["update_id"] + 1

    async def send_message(self, recipient_id, text, attachments, **kwargs):
        results = []

        async with self.api_messages_lock:
            if text:
                results.append(
                    await self.request(
                        "sendMessage",
                        {
                            "chat_id": recipient_id,
                            "text": str(text),
                            **kwargs,
                        },
                    )
                )
                await asyncio.sleep(self.api_messages_pause)

            for attachment in attachments or ():
                if not isinstance(attachment, Attachment):
                    raise ValueError(f'Unexpected attachment: "{attachment}"')

                if attachment.kind not in SUPPORTED_ATTACHMENT_KINDS:
                    raise ValueError(
                        f'Unsupported attachment kind: "{attachment.kind}"'
                    )

                method, data_field = SUPPORTED_ATTACHMENT_KINDS[attachment.kind]

                results.append(
                    await self.request(
                        method,
                        pick_by(
                            {
                                "chat_id": recipient_id,
                                data_field: attachment.id or attachment.content or None,
                                "caption": attachment.title
                                or (
                                    attachment.content[0]
                                    if attachment.content
                                    else None
                                ),
                            }
                        ),
                    )
                )

                await asyncio.sleep(self.api_messages_pause)

        return results

    async def on_shutdown(self, app):
        await self.client.aclose()



import asyncio
import logging
from itertools import groupby
from typing import Dict, List

from .backend import Backend
from .context import Context
from .plugin import Plugin
from .router import ListRouter, Router
from .storage import Storage
from .storages import MemoryStorage

# Find proper methods for different python versions
if hasattr(asyncio.Task, "all_tasks"):
    _all_tasks = asyncio.Task.all_tasks  # type: ignore
else:
    _all_tasks = asyncio.all_tasks

if hasattr(asyncio.Task, "current_task"):
    _current_task = asyncio.Task.current_task  # type: ignore
else:
    _current_task = asyncio.current_task


class Kutana:
    """
    Main class for kutana application

    Configuration:

    - '.prefixes' - prefixes for commands (default is [".", "/"])
    - '.ignore_initial_spaces' - ignore spaces after prefix (default is True)

    :ivar ~.config: Application's configuration
    """

    def __init__(
        self,
        concurrent_handlers_count=512,
    ):
        self._plugins: List[Plugin] = []
        self._backends: List[Backend] = []
        self._storages: Dict[str, Storage] = {"default": MemoryStorage()}
        self._root_router: Router

        self._hooks = {
            "start": [],
            "exception": [],
            "completion": [],
            "shutdown": [],
        }

        self._concurrent_handlers_count = concurrent_handlers_count
        self._semaphore = asyncio.Semaphore(value=concurrent_handlers_count)

        self.config = {
            "prefixes": ("/",),
            "mention_prefixes": ("", ","),
        }

    def _prepare_routers(self):
        source_routers = []

        for plugin in self._plugins:
            source_routers.extend(plugin._routers)

        sorted_source_routers = sorted(
            source_routers, reverse=True, key=lambda router: router.priority
        )

        root_router = ListRouter()

        for _, outer_group in groupby(
            sorted_source_routers, key=lambda router: router.priority
        ):
            for cls, inner_group in groupby(outer_group, key=type):
                if issubclass(cls, Router):
                    root_router.add_handler(cls.merge(inner_group))

        self._root_router = root_router

    async def _handle_event(self, event, *args, **kwargs):
        for handler in self._hooks[event]:
            try:
                await handler(*args, **kwargs)
            except Exception:
                logging.exception('Error while handling event "%s"', event)

    def add_storage(self, name, storage):
        """Add storage to the application (replaces previous if needed)."""
        if not isinstance(storage, Storage):
            raise ValueError(f"Provided value is not a storage: {storage}")
        self._storages[name] = storage

    @property
    def storages(self):
        return self._storages

    def add_plugin(self, plugin):
        """Add plugin to the application."""
        if not isinstance(plugin, Plugin):
            raise ValueError(f"Provided value is not a plugin: {plugin}")

        if plugin in self._plugins:
            raise RuntimeError("Plugin already added")

        plugin.app = self
        self._plugins.append(plugin)

        self._prepare_routers()

    @property
    def plugins(self):
        return self._plugins

    def add_backend(self, backend):
        """Add backend to the application."""
        if not isinstance(backend, Backend):
            raise ValueError(f"Provided value is not a backend: {backend}")

        if backend in self._backends:
            raise RuntimeError("Backend already added")

        self._backends.append(backend)

    @property
    def backends(self):
        return self._backends

    async def _run_wrapper(self):
        try:
            return await self._run()
        except KeyboardInterrupt:
            pass
        except Exception:
            logging.exception("Error while running application:")

    async def _run(self):
        logging.debug("Initiating storages")

        for storage in self._storages.values():
            await storage.init()

        logging.debug("Initiating hooks")
        for plugin in self._plugins:
            for event, handler in plugin._hooks:
                self._hooks[event].append(handler)

        logging.debug("Creating queue for acquired updates")
        queue = asyncio.Queue(maxsize=self._concurrent_handlers_count)

        logging.debug("Preparing backends and starting background updates acquiring")
        for backend in self._backends:
            await backend.on_start(self)
            asyncio.ensure_future(backend.acquire_updates(queue))

        logging.debug("Handling start event")
        await self._handle_event("start")

        logging.debug("Running processing loop")
        try:
            while True:
                await self._semaphore.acquire()

                update, backend = await queue.get()
                context = Context(self, update, backend)
                await backend.setup_context(context)

                task = asyncio.ensure_future(self._handle_update(context))
                task.add_done_callback(lambda _: self._semaphore.release())
        except asyncio.CancelledError:
            raise
        except Exception:
            self.stop()
            raise

    async def _handle_update(self, context):
        logging.debug("Processing update %s", context.update)

        try:
            await self._root_router.handle(context.update, context)
            await self._handle_event("completion", context)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            logging.exception("Exception while handling the update")
            await self._handle_event("exception", context, exc)

    async def _shutdown_wrapper(self):
        try:
            return await self._shutdown()
        except Exception:
            logging.exception("Error while shutting down application:")

    async def _shutdown(self):
        logging.info("Gracecfully shutting application down...")

        # Cancel everything
        tasks = []

        for task in _all_tasks():
            if task is not _current_task():
                task.cancel()
                tasks.append(task)

        await asyncio.gather(*tasks, return_exceptions=True)

        # Clean up
        tasks = []

        for backend in self._backends:
            tasks.append(backend.on_shutdown(self))

        tasks.append(self._handle_event("shutdown"))

        await asyncio.gather(*tasks, return_exceptions=True)

        # Stop loop
        asyncio.get_event_loop().stop()

    def run(self):
        """Run the application."""
        logging.info("Starting application...")

        try:
            asyncio.ensure_future(self._run_wrapper())
            asyncio.get_event_loop().run_forever()
        except KeyboardInterrupt:
            asyncio.ensure_future(self._shutdown_wrapper())
            asyncio.get_event_loop().run_forever()
        finally:
            asyncio.get_event_loop().close()
            logging.info("Stopped application")

    def stop(self):
        """Stop the application."""
        return asyncio.ensure_future(self._shutdown_wrapper())



class Backend:
    def get_identity(self):
        raise NotImplementedError

    async def on_start(self, app):
        pass

    async def on_shutdown(self, app):
        pass

    async def setup_context(self, context):
        pass

    async def acquire_updates(self, queue):
        raise NotImplementedError

    async def send_message(self, recipient_id, text=None, attachments=None, **kwargs):
        raise NotImplementedError

    async def request(self, method, kwargs):
        raise NotImplementedError



import io
from enum import Enum
from typing import Any, Awaitable, Callable, List, NamedTuple, Optional, Tuple, Union


class AttachmentKind(str, Enum):
    IMAGE = "image"
    DOCUMENT = "document"
    VIDEO = "video"
    AUDIO = "audio"
    VOICE = "voice"


class AttachmentContent(NamedTuple):
    name: str
    value: Union[io.IOBase, bytes]


class Attachment:
    def __init__(
        self,
        id: Optional[Union[str, int, float]] = None,
        kind: Optional[Union[AttachmentKind, str]] = None,
        content: Optional[
            Union[AttachmentContent, Tuple[str, Union[str, bytes]]]
        ] = None,
        title: Optional[str] = None,
        raw: Any = None,
        get_file: Optional[Callable[..., Awaitable]] = None,
    ):
        if not id and not content:
            raise ValueError("Attachment at least should have id or content")

        self.id = id
        self.kind = kind
        self.content = content  # named tuple of (name of content, content)
        self.title = title  # human-friendly description of attachment
        self.raw: Any = raw
        self._get_file = get_file
        self._file = None

    async def get_content(self):
        if self._file is not None:
            return self._file

        if self._get_file is None:
            raise ValueError("Can't get content for this file")

        file = await self._get_file()
        self._file = file
        return file


class RecipientKind(str, Enum):
    GROUP_CHAT = "group_chat"
    PRIVATE_CHAT = "private_chat"


class Message:
    def __init__(
        self,
        sender_id: str,
        recipient_id: str,
        recipient_kind: RecipientKind,
        text: str,
        attachments: List[Attachment],
        date: int,
        raw: Any,
    ):
        self.sender_id: str = sender_id
        self.recipient_id: str = recipient_id
        self.recipient_kind: RecipientKind = recipient_kind
        self.text: str = text
        self.attachments: List[Attachment] = attachments
        self.date: int = date
        self.raw: Any = raw
